<?php

namespace ZRay;

use Throwable;

/**
 * Class \ZRay\Composer
 *
 * Collects information about PHP autoloaded classes
 */
class Composer
{
    /**
     * @var array Class map data collection
     */
    protected array $classMap = [];
    /**
     * Collects information about class loads
     *
     * @param array $context
     * @param array $storage
     *
     * @return void
     */
    public function findFileWithExtensionExit(array $context, array &$storage): void
    {
        try {
            if ($context['returnValue'] === null) {
                // failed lookup, so class was not in Composer class map
                return;
            }

            $filename = str_replace(['/', "\\"], DIRECTORY_SEPARATOR, $context['returnValue']);
            // try to remove everything below "vendor" directory
            $splitPos = strpos($filename, 'composer' . DIRECTORY_SEPARATOR . '..');
            if ($splitPos) {
                $filename = 'vendor' . substr($filename, $splitPos + 11);
            }
            $this->classMap[] = [
                'class' => $context['functionArgs'][0],
                'filename' => $filename
            ];
        } catch (Throwable $e) {
            error_log("ZRay Composer plugin failed: " . $e->getMessage());
        }
    }

    /**
     * Collects information about all currently installed packages
     *
     * @param array $context
     * @param array $storage
     *
     * @return void
     */
    public function registerExit(array $context, array &$storage): void
    {
        try {
            if (!isset($context['calledFromFile'])) {
                return;
            }
            $composerDir = dirname($context['calledFromFile']);
            $jsonFile = $composerDir . '/installed.json';

            if (!file_exists($jsonFile) || !is_readable($jsonFile)) {
                return;
            }
            $entries = [];

            $json = file_get_contents($jsonFile);
            $data = json_decode($json);

            /*
             * Composer 1.x - output is already an array
             * Composer 2.x - output is an object; the array we want is in the 'packages' property
             */
            if (is_object($data) && isset($data->packages)) {
                $data = $data->packages;
                $entries['Composer : 2'] = [];
            } else {
                $entries['Composer : 1'] = [];
            }

            foreach ($data as $package) {
                if (!is_object($package)) {
                    continue;
                }
                $entry = [];
                if (!empty($package->source->url)) {
                    $entry['source'] = $package->source->url;
                }
                if (!empty($package->description)) {
                    $entry['description'] = $package->description;
                }
                if (!empty($package->require)) {
                    $entry['requires'] = (array)$package->require;
                }
                if (!empty($package->authors)) {
                    $entry['authors'] = [];
                    foreach ($package->authors as $author) {
                        if (!empty($author->name)) {
                            $entry['authors'][$author->name] = $author->email ?? '';
                        }
                    }
                }
                if (!empty($package->homepage)) {
                    $entry['homepage'] = $package->homepage;
                }
                $entries[$package->name . ' : ' . $package->version] = $entry;
            }
            $storage['packages'][] = $entries;
        } catch (Throwable $e) {
            error_log("ZRay Composer plugin failed: " . $e->getMessage());
        }
    }

    public function collect($context, &$storage)
    {
        $storage['classMap'][] = $this->classMap;
    }

    public static function shutdown(): bool
    {
        return true;
    }
}
