<?php

/**
 * Sake
 *
 * @link      http://github.com/sandrokeil/Z-Ray-Doctrine2 for the canonical source repository
 * @copyright Copyright (c) 2015 Sandro Keil
 * @license   http://github.com/sandrokeil/Z-Ray-Doctrine2/blob/master/LICENSE.txt New BSD License
 */

namespace ZRay;

use Doctrine\Common\Cache\Cache;
use Doctrine\DBAL\Connection;
use Doctrine\DBAL\Statement;
use Doctrine\ORM\EntityManager;
use Doctrine\ORM\Mapping\ClassMetadata;
use Psr\Cache\CacheItemPoolInterface;

/**
 * Class Doctrine
 *
 * Collects information about Doctrine ORM and DBAL.
 */
class Doctrine
{
    /**
     * Entities
     *
     * @var array
     */
    protected array $entities = [];

    /**
     * Queries
     *
     * @var array
     */
    protected array $queries = [];

    /**
     * Query count
     *
     * @var int
     */
    protected int $queryNumber = 0;

    /**
     * Cache settings
     */
    protected array $cache
        = [
            'metadata' => [
                'name' => 'Metadata Cache',
                'status' => 'none used'
            ],
            'query' => [
                'name' => 'Query Cache',
                'status' => 'none used'
            ],
            'result' => [
                'name' => 'Result Cache',
                'status' => 'none used'
            ],
            'hydration' => [
                'name' => 'Hydration Cache',
                'status' => 'none used'
            ],
            'secondLevel' => [
                'name' => 'Second Level Cache',
                'status' => 'none used'
            ],
        ];

    /**
     * Events
     */
    protected array $events = [];

    /**
     * Collects queries from \Doctrine\ORM\Persisters\Entity\BasicEntityPersister
     *
     * @param array $context
     * @param array $storage
     */
    public function cache(array $context, array &$storage): void
    {
        if (!empty($context['functionArgs'][0]) && $context['functionArgs'][0] instanceof CacheItemPoolInterface) {
            $class = get_class($context['functionArgs'][0]);
        } else {
            if (is_object($context['returnValue'])) {
                $class = get_class($context['returnValue']);
            } else {
                return;
            }
        }

        $status = str_replace('Doctrine\Common\Cache', '', $class) . ' used';

        switch ($context['functionName']) {
            case 'Doctrine\ORM\Configuration::setMetadataCache':
                $this->cache['metadata']['status'] = $status;
                break;
            case 'Doctrine\ORM\Configuration::setSecondLevelCacheEnabled':
                $this->cache['secondLevel']['status'] = $status;
                break;
            case 'Doctrine\DBAL\Configuration::setResultCache':
                $this->cache['result']['status'] = $status;
                break;
            case 'Doctrine\ORM\Configuration::setQueryCache':
                $this->cache['query']['status'] = $status;
                break;
            case 'Doctrine\ORM\Configuration::setHydrationCache':
                $this->cache['hydration']['status'] = $status;
                break;
            default:
                break;
        }
    }

    /**
     * Collects all queries from Doctrine\DBAL\Connection
     *
     * @param array $context
     * @param array $storage
     */
    public function connection(array $context, array &$storage): void
    {
        $query = $context['locals']['sql'] ?? '';

        if (!$query) {
            return;
        }

        $queryId = $this->getQueryId($query);

        if (!array_key_exists($queryId, $this->queries)) {
            $this->queryNumber++;

            $this->queries[$queryId] = [
                'id' => $this->queryNumber,
                'query' => $query,
                'count' => 0,
                'cached' => 0,
            ];
        }

        if ($context['functionName'] === 'Doctrine\DBAL\Connection::executeCacheQuery') {
            $this->queries[$queryId]['cached']++;
            return;
        }

        $this->queries[$queryId]['count']++;
    }

    /**
     * Collects all SQL transaction calls
     *
     * @param array $context
     * @param array $storage
     */
    public function transaction(array $context, array &$storage): void
    {
        [, $query] = explode('::', $context['functionName']);
        $query = str_replace('beginT', 'BEGIN T', $query);
        $this->queryNumber++;

        $this->queries[$query . $this->queryNumber] = [
            'id' => $this->queryNumber,
            'query' => strtoupper($query),
            'count' => 1,
            'cached' => 0,
        ];
    }

    /**
     * Collects queries from Doctrine\DBAL\Statement
     *
     * @param array $context
     * @param array $storage
     */
    public function statement(array $context, array &$storage): void
    {

        $the = $context['this'] ?? '';
        if (!($the instanceof Statement)) {
            return;
        }

        $query = $the->sql;
        if (!$query) {
            return;
        }

        $this->queryNumber++;
        $queryId = $this->getQueryId($query);

        if (!isset($this->queries[$queryId])) {
            $this->queries[$queryId] = [
                'id' => $this->queryNumber,
                'query' => $query,
                'count' => 1,
                'cached' => 0,
            ];
        } else {
            $this->queries[$queryId]['count']++;
        }
    }

    /**
     * Collects entity usage from unit of work.
     *
     * @param array $context
     * @param array $storage
     */
    public function unitOfWork(array $context, array &$storage): void
    {
        // dont break z-ray
        if (empty($context['functionArgs'][0]) || empty($context['returnValue'])) {
            return;
        }

        $class = $context['functionArgs'][0];
        $hash = spl_object_hash($context['returnValue']);

        if (!isset($this->entities[$hash])) {
            $this->entities[$hash] = ['class' => $class, 'unique' => 1, 'ref' => 0];
        } else {
            $this->entities[$hash]['ref']++;
        }
    }

    /**
     * Collects entity usage from unit of work.
     *
     * @param array $context
     * @param array $storage
     */
    public function entityMapping(array $context, array &$storage): void
    {
        // dont break z-ray
        if (empty($context['functionArgs'][0]) || !($context['functionArgs'][0] instanceof EntityManager)) {
            return;
        }

        /* @var $em EntityManager */
        $em = $context['functionArgs'][0];

        $allMetadata = $em->getMetadataFactory()->getAllMetadata();

        if (empty($allMetadata)) {
            return;
        }
        /* @var $metadata ClassMetadata */
        foreach ($allMetadata as $metadata) {
            $this->entities[] = [
                'class' => $metadata->getName(),
                'unique' => 0,
                'ref' => 0,
            ];
        }
    }

    /**
     * Collects events from Doctrine EventManager
     *
     * @param array $context
     * @param array $storage
     */
    public function eventManagerDispatch(array $context, array &$storage)
    {
        // dont break z-ray
        if (empty($context['functionArgs'][0])) {
            return;
        }

        $eventName = $context['functionArgs'][0];

        if (empty($this->events[$eventName])) {
            $this->events[$eventName] = [
                'name' => $eventName,
                'count' => 0,
                'listeners' => [],
            ];
        }

        $this->events[$eventName]['count']++;

        if (empty($context['locals']['listener'])) {
            return;
        }
        $listener = get_class($context['locals']['listener']);
        $this->events[$eventName]['listeners'][$listener] = true;
    }

    /**
     * Collects listener from Doctrine EventManager
     *
     * @param array $context
     * @param array $storage
     */
    public function eventManagerAddListener(array $context, array &$storage): void
    {
        // dont break z-ray
        if (empty($context['functionArgs'][0]) || empty($context['functionArgs'][1])) {
            return;
        }

        $events = (array)$context['functionArgs'][0];

        $listener = $context['functionArgs'][1];

        if (is_object($listener)) {
            $listener = get_class($listener);
        }

        foreach ($events as $event) {
            if (empty($this->events[$event])) {
                $this->events[$event] = [
                    'name' => $event,
                    'count' => 0,
                    'listeners' => [$listener => true],
                ];
            } else {
                $this->events[$event]['listeners'][$listener] = true;
            }
        }
    }

    /**
     * Collects all data from other functions to display information in Z-Ray
     *
     * @param array $context
     * @param array $storage
     */
    public function collectAllData(array $context, array &$storage): void
    {
        // collect entities
        if (empty($this->queries) && empty($this->entities)) {
            // no queries and entities - no Doctrine usage
            return;
        }

        foreach ($this->entities as $data) {
            if (!isset($storage['entities'][$data['class']])) {
                $storage['entities'][$data['class']] = [
                    'entity' => $data['class'],
                    'unique_entities' => $data['unique'],
                    'referenced_entities' => $data['ref'],
                ];
                continue;
            }
            $storage['entities'][$data['class']]['unique_entities'] += $data['unique'];
            $storage['entities'][$data['class']]['referenced_entities'] += $data['ref'];
        }

        // collect queries
        foreach ($this->queries as $query => $data) {
            $storage['queries'][$query] = $data;
        }

        // collect cache
        $storage['cache'] = $this->cache;

        $events = [
            'onClear',
            'onFlush',
            'postFlush',
            'postLoad',
            'postPersist',
            'postRemove',
            'postUpdate',
            'preFlush',
            'prePersist',
            'preRemove',
            'preUpdate',
            'loadClassMetadata',
            'onClassMetadataNotFound',
        ];

        foreach ($events as $event) {
            $storage['events'][$event] = [
                'name' => $event,
                'count' => 0,
                'listeners' => '',
            ];
        }

        foreach ($this->events as $event => $data) {
            $storage['events'][$event] = [
                'name' => $data['name'],
                'count' => $data['count'],
                'listeners' => array_keys($data['listeners']),
            ];
        }
    }

    /**
     * Only for listen to the php shutdown event to collect all data
     *
     * @see \ZRay\Doctrine::collectAllData
     */
    public function shutdown()
    {
    }

    /**
     * Returns a unique query id
     *
     * @param string $query
     *
     * @return string Unique id
     */
    private function getQueryId(string $query): string
    {
        return md5($query);
    }
}
